@@ -3624,7 +3624,7 @@ const nghttp2_huff_decode huff_decode_table[][16] = {
     {0, 0x03, 253},
     {0, 0x03, 254},
     {0, 0x03, 255},
-    {-1, 0x00, 0},
+    {0, 0x04, 0},
     {189, 0x00, 0},
     {191, 0x00, 0},
     {192, 0x00, 0},
@@ -3772,8 +3772,8 @@ const nghttp2_huff_decode huff_decode_table[][16] = {
     {14, 0x03, 254},
     {1, 0x02, 255},
     {14, 0x03, 255},
-    {-1, 0x00, 0},
-    {-1, 0x00, 0},
+    {0, 0x04, 0},
+    {0, 0x04, 0},
     {0, 0x03, 0},
     {0, 0x03, 1},
     {0, 0x03, 2},
@@ -3840,10 +3840,10 @@ const nghttp2_huff_decode huff_decode_table[][16] = {
   },
   /* 187 */
   {
-    {-1, 0x00, 0},
-    {-1, 0x00, 0},
-    {-1, 0x00, 0},
-    {-1, 0x00, 0},
+    {0, 0x04, 0},
+    {0, 0x04, 0},
+    {0, 0x04, 0},
+    {0, 0x04, 0},
     {1, 0x02, 0},
     {14, 0x03, 0},
     {1, 0x02, 1},
@@ -3859,14 +3859,14 @@ const nghttp2_huff_decode huff_decode_table[][16] = {
   },
   /* 188 */
   {
-    {-1, 0x00, 0},
-    {-1, 0x00, 0},
-    {-1, 0x00, 0},
-    {-1, 0x00, 0},
-    {-1, 0x00, 0},
-    {-1, 0x00, 0},
-    {-1, 0x00, 0},
-    {-1, 0x00, 0},
+    {0, 0x04, 0},
+    {0, 0x04, 0},
+    {0, 0x04, 0},
+    {0, 0x04, 0},
+    {0, 0x04, 0},
+    {0, 0x04, 0},
+    {0, 0x04, 0},
+    {0, 0x04, 0},
     {2, 0x02, 0},
     {6, 0x02, 0},
     {15, 0x02, 0},
