@@ -173,7 +173,7 @@ void test_nghttp2_frame_pack_headers_frame_too_large(void)
   nghttp2_bufs bufs;
   nghttp2_nv *nva;
   ssize_t nvlen;
-  size_t big_vallen = NGHTTP2_HD_MAX_VALUE;
+  size_t big_vallen = NGHTTP2_HD_MAX_NV;
   nghttp2_nv big_hds[16];
   size_t big_hdslen = ARRLEN(big_hds);
   size_t i;
